(function (GCN) {

	'use strict';

	/**
	 * Maps contracts that were fetched via the Rest API into a flat hash map.
	 *
	 * @param {object<string, object>} categories
	 * @return {object<string, object>}
	 */
	function mapConstructs(categories) {
		var groupName;
		var group;
		var constructName;
		var construct;
		var map = {};

		for (groupName in categories) {
			if (categories.hasOwnProperty(groupName)) {
				group = categories[groupName].constructs;

				for (constructName in group) {
					if (group.hasOwnProperty(constructName)) {
						construct = group[constructName];
						map[constructName] = construct;
					}
				}
			}
		}

		return map;
	}

	/**
	 * Node object.
	 *
	 * @name NodeAPI
	 * @class
	 * @augments Chainback
	 */
	var NodeAPI = GCN.defineChainback({
		/** @lends NodeAPI */

		__chainbacktype__: 'NodeAPI',
		_extends: GCN.ContentObjectAPI,
		_type: 'folder',

		_data: {
			folderId: null
		},

		/**
		 * @protected
		 * @type {object<string, number} Constructs for this node are cached
		 *                               here so that we only need to fetch
		 *                               this once.
		 */
		_constructs: null,

		/**
		 * Retrieves a list of constructs and constructs categories that are
		 * assigned to this node and passes it as the only argument into the
		 * the `success()' callback.
		 *
		 * @param {function(Array.<object>)=} success Callback to receive an
		 *                                            array of constructs.
		 * @param {function(GCNError):boolean=} error Custom error handler.
		 * @return Returns the constructs / categories
		 * @throws INVALID_ARGUMENTS
		 */
		constructs: function (success, error) {
			if (!success) {
				GCN.error('INVALID_ARGUMENTS', 'the `constructs()\' method ' +
					'requires at least a success callback to be given');
			}

			if (this._constructs) {
				success(this._constructs);
			} else {
				var that = this;

				this._continueWith(function (child) {
					that._data.id = child._data.nodeId;

					that.constructCategories(function (categories) {
						that._constructs = mapConstructs(categories);
						success(that._constructs);
					}, error);
				}, error);
			}
		},

		/**
		 * Removes this node object.
		 *
		 * @param {function()=} success Callback.
		 * @param {function(GCNError):boolean=} error Custom error handler.
		 * @param {function} success callback
		 */
		remove: function (success, error) {

		},

		/**
		 * @FIXME: Is it really possible to save changes to a node?  If not,
		 *         then we should not surface this method.
		 */
		save: function () {},

		/**
		 * Retreives the top-level folders of this node's root folder.
		 *
		 * @param {function(FolderAPI)=} success
		 * @param {function(GCNError):boolean=} error Custom error handler.
		 */
		'!folders': function (success, error) {
			return this.folder(null, error).folders(success, error);
		},

		/**
		 * Helper method that will load the constructs of this node.
		 *
		 * @private
		 * @this {NodeAPI}
		 * @param {function(Array.<object>)} success callback
		 * @param {function(GCNError):boolean=} error callback
		 */
		constructCategories: function (success, error) {
			this._authAjax({
				url   : GCN.settings.BACKEND_PATH + '/rest/construct/load.json',
				type  : 'GET',
				error : function (xhr, status, msg) {
					GCN.handleHttpError(xhr, msg, error);
				},
				success: function (response) {
					if (GCN.getResponseCode(response) === 'OK') {
						success(response.constructCategories);
					} else {
						GCN.handleResponseError(response, error);
					}
				}
			});
		}

	});

	GCN.node = GCN.exposeAPI(NodeAPI);
	GCN.NodeAPI = NodeAPI;

}(GCN));
